/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/***************************************************************************
 *            decoder.cc
 *
 *  Sat Feb 19 17:05:43 CET 2005
 *  Copyright  2005 Bent Bisballe
 *  deva@aasimon.org
 ****************************************************************************/

/*
 * Originally from:
 * RTVideoRec Realtime video recoder and encoder for Linux
 *
 * Copyright (C) 2004  B. Stultiens
 * Copyright (C) 2004  Koen Otter and Glenn van der Meyden
 */

/*
 *    This file is part of MIaV.
 *
 *    MIaV is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    MIaV is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with MIaV; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */

/*
 * $Id: decoder.cc,v 1.33 2005/06/16 21:28:57 deva Exp $
 */

/*
 * $Log: decoder.cc,v $
 * Revision 1.33  2005/06/16 21:28:57  deva
 * Rewrote thread object
 * Fixed bug in mov_encoder (pushed read_sem too many times, whihc lead to
 * growing server queue)
 *
 * Revision 1.32  2005/06/09 11:00:03  deva
 * Added daemon code, and cleaned up using -Wall and -Werror
 *
 * Revision 1.31  2005/06/02 20:45:01  deva
 *
 * Added clear button
 * Optimized the frame handling a little (very little!).
 *
 * Revision 1.30  2005/06/02 15:03:23  deva
 *
 * Fixed crash in network.cc if socket not connected.
 * Added option to skop ecery second frame in player
 *
 * Revision 1.29  2005/05/25 13:11:42  deva
 *
 * Made unfreeze close connection, when no recording is done.
 *
 * Revision 1.28  2005/05/16 16:00:56  deva
 *
 * Lots of stuff!
 *
 * Revision 1.27  2005/05/03 08:31:59  deva
 * Removed the error object, and replaced it with a more generic info object.
 *
 * Revision 1.26  2005/05/02 14:06:55  deva
 * Added unfreeze to decoder (called from camera).
 *
 * Revision 1.25  2005/05/02 10:52:46  deva
 *
 * Fixed bug invoking infinite loop, when snapshot is requested.
 *
 * Revision 1.24  2005/05/02 10:35:23  deva
 * Fixed wrongly showed snapshot thumbnails.
 *
 * Revision 1.23  2005/05/02 09:58:43  deva
 *
 * Fixed initial values of the stae bools.
 *
 * Revision 1.22  2005/05/02 09:50:22  deva
 * Rewrote freeze, shoot and record flags, from encoder to frame.
 *
 * Revision 1.21  2005/05/01 09:56:26  deva
 * Added Id and Log tags to all files
 */

#include <config.h>
#ifdef USE_GUI
/*
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
*/

#include "miav_config.h"

#include <time.h>

// Use libdv
#include <libdv/dv.h>
#include <libdv/dv_types.h>

#include <SDL/SDL.h>

#include "dv1394.h"
#include "dv.h"

#include "decoder.h"
#include "debug.h"

Decoder::Decoder(Info *ginfo,
                 sem_t *gencode_sem,
                 sem_t *gplayer_sem,
                 Queue<Frame> *gencode_queue,
                 Queue<Frame> *gplayer_queue,
                 pthread_mutex_t *gmutex,
                 volatile int *grunning)
{
  info = ginfo;

  encode_sem = gencode_sem;
  player_sem = gplayer_sem;
  encode_queue = gencode_queue;
  player_queue = gplayer_queue;
  mutex = gmutex;
  running = grunning;

  b_shoot = false;
  b_freeze = false;
  b_record = false; // Initially no recording is done.

  pthread_mutex_init (&shot_mutex, NULL);
  shot = NULL;
}

Decoder::~Decoder()
{
  pthread_mutex_destroy(&shot_mutex);
}

void Decoder::decode()
{
  bool local_shoot;
  int local_freeze;
  bool local_record = false;
  bool old_record;

  bool skip_frames = config->readInt("player_skip_frames");

  dv1394 dv_stream = dv1394(info); // Use default port and channel.

  while(*running) {
    uint8_t *ptr;
    SDL_Event user_event;
    
    // Read a dvframe
    ptr = dv_stream.readFrame();
    if(!ptr) return; // No frame read. (Due to firewire error)

    old_record = local_record;
    local_shoot = b_shoot;
    b_shoot = false;
    local_freeze = b_freeze;
    b_freeze = false;
    local_record = b_record;

    if(local_shoot) {
      pthread_mutex_lock(&shot_mutex);
      if(!shot) shot = new Frame(ptr, DVPACKAGE_SIZE);
      pthread_mutex_unlock(&shot_mutex);
    }
    
    if(local_freeze == 1) {
      pthread_mutex_lock(&shot_mutex);
      if(shot) delete shot;
      shot = new Frame(ptr, DVPACKAGE_SIZE);
      pthread_mutex_unlock(&shot_mutex);
    }

    static int showframe = 1;
    if(skip_frames != 0) showframe = 1 - showframe;
    if(showframe) {
      Frame *pframe = new Frame(ptr, DVPACKAGE_SIZE);

      pframe->shoot = local_shoot;
      pframe->freeze = local_freeze;
      pframe->record = local_record;

      player_queue->push(pframe);
    
      // Create and send SDL event.
      user_event.type = SDL_USEREVENT;
      user_event.user.code = 0;
      user_event.user.data1 = NULL;
      user_event.user.data2 = NULL;
      SDL_PushEvent(&user_event);
    }

    if(local_record | (local_record != old_record) | local_shoot | local_freeze) {
      Frame *eframe = new Frame(NULL, 0);
      eframe->data = ptr;
      eframe->size = DVPACKAGE_SIZE;

      eframe->shoot = local_shoot;
      eframe->freeze = local_freeze;
      eframe->record = local_record;
    
      encode_queue->push(eframe);
      
      sem_post(encode_sem);
    } else {
      free(ptr);
    }
  }

  // Kick the others so they wake up with empty queues
  sem_post(encode_sem);
}

void Decoder::thread_main() {
  decode();
  fprintf(stderr, "Decoder thread stopped.\n"); fflush(stderr);
}

/*
 * Set freeze bit on current frame.
 */
void Decoder::freeze()
{  
  b_freeze = 1;
}

/*
 * Remove frozen frame.
 */
void Decoder::unfreeze()
{  
  b_freeze = -1;

  pthread_mutex_lock(&shot_mutex);
  delete shot;
  shot = NULL;
  pthread_mutex_unlock(&shot_mutex);
}

/*
 * Set shoot bit on current frame.
 */
void Decoder::shoot(unsigned char *rgb)
{
  struct timespec ts;

  b_shoot = true;

  // Wait for shot to be taken
  while(1) {
    pthread_mutex_lock(&shot_mutex);
    if(shot) {
      getScreenshot(shot, rgb);
      delete shot;
      shot = NULL;
      pthread_mutex_unlock(&shot_mutex);
      return;
    }
    pthread_mutex_unlock(&shot_mutex);

    ts.tv_sec = 0;
    ts.tv_nsec = 100000000L;	// 100ms
    nanosleep(&ts, NULL);
  }
}

/*
 * Set the record bit to true in all following frames.
 */
void Decoder::start()
{
  b_record = true;
}

/*
 * Set the record bit to false in all following frames.
 */
void Decoder::stop(n_savestate save)
{
  b_record = false;
}

void Decoder::getScreenshot(Frame *frame, unsigned char *rgb)
{
  unsigned char *pixels[3];
  int pitches[3];

  pixels[ 0 ] = rgb;
  pixels[ 1 ] = NULL;
  pixels[ 2 ] = NULL;

  pitches[ 0 ] = 720 * 4;
  pitches[ 1 ] = 0;
  pitches[ 2 ] = 0;
  
	dv_decoder_t *decoder = dv_decoder_new(FALSE/*this value is unused*/, FALSE, FALSE);
  decoder->quality = DV_QUALITY_BEST;

  dv_parse_header(decoder, frame->data);
  
  decoder->system = e_dv_system_625_50;  // PAL lines, PAL framerate
  decoder->sampling = e_dv_sample_422;  // 4 bytes y, 2 bytes u, 2 bytes v
  decoder->std = e_dv_std_iec_61834;
  decoder->num_dif_seqs = 12;
  
  // libdv img decode to rgb
  dv_decode_full_frame(decoder,
                       frame->data,
                       e_dv_color_bgr0,
                       pixels,
                       pitches);
  
  dv_decoder_free(decoder);
}


#endif /*USE_GUI*/
